from saga.Repository import Repository
from saga.operations.get_remote import get_remote

def pull(repository: Repository):
    url = f"{repository.remote_repository}/cli/pull/{repository.base_directory.name}"
    get_remote(repository.remote_repository)
    print(f"Got remote from {url}")



from pathlib import Path
from saga.file_types.binary_file import parse_binary_file, write_binary_file
from saga.file_types.text_file import parse_text_file, write_text_file
from saga.file_types.csv_file import parse_csv_file, write_csv_file
from saga.file_types.excel_file import parse_excel_file, write_excel_file


def is_csv(file_path: Path):
    return file_path.suffix == ".csv"


TEXT_ENDINGS = {
    ".txt",  # just text
    ".py",  # python
    ".c",  # c source
    ".cc",  # c++ source
    ".java",  # java source
    ".js",  # js source
    ".php"  # php source
}


def is_text(file_path: Path):
    return file_path.suffix in TEXT_ENDINGS


def is_excel(file_path: Path):
    return file_path.suffix == ".xlsx"


def parse_file(file_id, file_name, file_path):
    if is_text(file_path):
        return parse_text_file(file_id, file_name, file_path)
    elif is_excel(file_path):
        return parse_excel_file(file_id, file_name, file_path)
    elif is_csv(file_path):
        return parse_csv_file(file_id, file_name, file_path)
    else:
        return parse_binary_file(file_id, file_name, file_path)


def write_file(file):
    if file.file_type == "csv":
        write_csv_file(file)
    elif file.file_type == "text":
        write_text_file(file)
    elif file.file_type == "binary":
        write_binary_file(file)
    elif file.file_type == "excel":
        write_excel_file(file)
    else:
        print("Error: file type {} does not exist".format(file.file_type))



from saga.path_utils import changed_files
from saga.Repository import Repository
from saga.CommitGraph import CommitGraph
from saga.file_types.file_utils import parse_file, write_file
from saga.operations.add import add
from saga.operations.commit import commit


def fast_forward(repository: Repository, other_branch: str):
    # Get the head commit on the other branch
    other_branch_head_hash = repository.head_commit_from_branch(other_branch)
    if other_branch_head_hash is None:
        print(f"Error: cannot fast forward as {other_branch} has no head")
        exit(1)

    other_branch_head_commit = repository.get_commit(other_branch_head_hash)
    repository.update_head_commit(other_branch_head_commit)
    repository.restore_state(other_branch_head_commit.state_hash)
    print(f"Fast-forward branch {repository.head} to branch {other_branch}")


def merge(repository: Repository, other_branch: str):
    """
    Merges current head with a other_branch
    """
    if other_branch not in repository.branches:
        print(
            f"Error: cannot merge branch \
            {other_branch} as it does not exist"
        )
        return

    # get the two commit hashes to find the common ancestor of
    commit_hash_1 = repository.head_commit_hash
    commit_hash_2 = repository.head_commit_from_branch(other_branch)

    # get their least common ancestor commit hashes
    commit_graph = CommitGraph(repository)
    need_merge, lcas = commit_graph.least_common_ancestors(
        commit_hash_1,
        commit_hash_2
    )

    if not need_merge:
        assert len(lcas) == 1
        if lcas.pop() == commit_hash_1:
            fast_forward(repository, other_branch)
        else:
            print(f"Branch {other_branch} is a subbranch of {repository.head}")
    else:
        # for now, we are gonna assume there is one lca
        lca = lcas.pop()
        state_dir_old = repository.state_directory / \
            repository.get_commit(lca).state_hash
        state_dir_newA = repository.state_directory / repository.state_hash
        state_dir_newB = repository.state_directory / \
            repository.get_commit(commit_hash_2).state_hash

        removed_paths_A, changed_paths_A, inserted_paths_A = changed_files(
            state_dir_old,
            state_dir_newA
        )
        removed_paths_B, changed_paths_B, inserted_paths_B = changed_files(
            state_dir_old,
            state_dir_newB
        )

        possibly_mergable = set()
        for path in removed_paths_A.intersection(changed_paths_B):
            print(
                f"Confict as {path} was removed in \
                {repository.head} and modified in {other_branch}"
            )
        for path in removed_paths_B.intersection(changed_paths_A):
            print(
                f"Confict as {path} was removed in \
                {repository.head} and modified in {other_branch}"
            )
        for path in changed_paths_A.intersection(changed_paths_B):
            possibly_mergable.add(path)
        for path in inserted_paths_A.intersection(inserted_paths_B):
            print(
                f"Conflict as {path} was inserted in \
                both {repository.head} and {other_branch}"
            )

        # we only write the files if there are no merge conflicts in any
        files_to_write = []
        for path in possibly_mergable:
            file_o = parse_file("TEMP", path, state_dir_old / path)
            file_a = parse_file("TEMP", path, state_dir_newA / path)
            file_b = parse_file("TEMP", path, state_dir_newB / path)

            merge_file = file_o.merge(file_a, file_b, path)
            if merge_file is None:
                print(f"Merge conflict for file {file_o.file_name}")
                return
            else:
                print(f"Successfully merged file {file_o.file_name}")
                files_to_write.append(merge_file)
        for f in files_to_write:
            write_file(f)
            add(repository, f.file_path)
        commit(
            repository,
            f"Merged {other_branch} into branch {repository.head}",
            allow_empty=True
        )



from shutil import copyfile
from saga.Repository import Repository


def create_branch(repository: Repository, branch_name: str):
    """
    If branch_name does not exist, will create a new branch with this name,
    branching off from the current commit on head
    """
    if (repository.branch_directory / branch_name).exists():
        print("Error: branch {} already exists".format(branch_name))

    head = repository.branch_directory / repository.head
    copyfile(head, repository.branch_directory / branch_name)



from saga.operations.commit import commit
from saga.commands.utils import get_saga_repo


def command_commit(args):
    saga_repo = get_saga_repo()
    if args.m is None:
        args.m = input("Please enter a commit message: ")
    commit(saga_repo, args.m, allow_empty=args.allow_empty)



import os
import hashlib
import requests
import shutil
from typing import Optional, List
from pathlib import Path
from os.path import join
from saga.Commit import Commit
from saga.path_utils import (
    copy_dir_to_dir,
    relative_paths_in_dir,
    changed_files
)
from saga.file_types.file_utils import parse_file


class Repository(object):

    def __init__(self, directory: Path):
        self.base_directory = directory

    def debug(self):
        for branch in self.branches:
            commit_hash = self.head_commit_from_branch(branch)
            string = ""
            while commit_hash:
                string = f"{commit_hash} - {string}"
                commit = self.get_commit(commit_hash)
                if commit is None or not any(commit.parent_commit_hashes):
                    commit_hash = None
                else:
                    commit_hash = commit.parent_commit_hashes[0]
            print(f"{branch}: {string}")

    @property
    def saga_directory(self) -> Path:
        # TODO: move all of these to paths
        return self.base_directory / ".saga/"

    @property
    def commit_directory(self) -> Path:
        return self.saga_directory / "commits"

    @property
    def state_directory(self) -> Path:
        return self.saga_directory / "states"

    @property
    def index_directory(self) -> Path:
        return self.saga_directory / "index"

    @property
    def branch_directory(self) -> Path:
        return self.saga_directory / "branches"

    @property
    def branches(self) -> List[str]:
        return [
            str(path.parts[-1]) for path in self.branch_directory.iterdir()
        ]

    @property
    def head_location(self) -> Path:
        return self.saga_directory / "head"

    @property
    def remote_location(self) -> Path:
        return self.saga_directory / "remote"

    @property
    def head(self) -> str:
        if self.head_location.exists():
            with open(self.head_location, "r") as f:
                return f.readline()
        else:
            with open(self.head_location, "w+") as f:
                f.write("master")
            return "master"

    @property
    def index_hash(self):
        """
        Returns the hash of a list of the files + contents, currently
        in the index folder.
        """
        file_hashes = []
        # we sort to have the same files every time
        for file_name in sorted(relative_paths_in_dir(self.index_directory)):
            index_file_name = join(self.index_directory, file_name)
            if os.path.isfile(index_file_name):
                f = open(index_file_name, "rb")
                file_bytes = f.read()
                m = hashlib.sha256()
                # we encode the file contents
                # and we encode the file path
                m.update(file_bytes)
                m.update(file_name.encode("utf-8"))
                file_hashes.append(m.hexdigest())
        m = hashlib.sha256()
        # TODO: this should be a merkle tree eventually
        # (or maybe an merkel-mountain-range), to reap the benefits
        m.update(",".join(file_hashes).encode('utf-8'))
        return m.hexdigest()

    @property
    def head_commit_hash(self):
        """
        Returns the commit hash on HEAD
        """
        return self.head_commit_from_branch(self.head)

    @property
    def remote_repository(self):
        """
        Returns the URL of the currently tracked remote repository
        """
        if self.remote_location.exists():
            with open(self.remote_location, "r") as f:
                return f.readline()
        else:
            with open(self.remote_location, "w+") as f:
                f.write("http://localhost:3000")
            return "http://localhost:3000"

    def head_commit_from_branch(self, branch_name: str) -> Optional[str]:
        branch_file = self.branch_directory / branch_name
        if not branch_file.exists():
            return None
        with open(branch_file, 'r') as f:
            return f.readline()

    @property
    def state_hash(self):
        """
        Returns the state hash of the most recent commit
        """
        head_commit = self.get_commit(self.head_commit_hash)
        return head_commit.state_hash

    @property
    def uncommited_in_index(self):
        index_state_hash = self.index_hash
        commit_state_hash = self.state_hash
        return index_state_hash != commit_state_hash

    def get_commit(self, commit_hash: str) -> Optional[Commit]:
        with (self.commit_directory / commit_hash).open("rb") as f:
            return Commit.from_bytes(f.read())
        return None

    def add_commit(
            self,
            state_hash: str,
            previous_commit_hashes: List[str],
            commit_message: str
            ) -> Optional[Commit]:
        """
        Adds a commit to the head branch
        """
        commit = Commit(state_hash, previous_commit_hashes, commit_message)
        commit_hash = commit.hash
        commit_bytes = commit.to_bytes()
        with open(self.commit_directory / commit_hash, "wb+") as f:
            f.write(commit_bytes)
        return commit

    def update_head_commit(self, commit: Commit):
        head = self.branch_directory / self.head
        with open(head, "w") as f:
            f.write(commit.hash)

    def set_head(self, branch_name: str):
        with open(self.head_location, "w") as f:
            f.write(branch_name)

    def restore_state(self, state_hash: str):
        # copy the state directory to the current directory
        copy_dir_to_dir(join(".saga/states", state_hash), ".")
        # remove all the files that shouldn't be in this state
        inserted, changed, removed = changed_files(
            ".",
            join(".saga/states", state_hash)
        )
        for path in inserted:
            if os.path.isdir(path):
                os.rmdir(path)
            else:
                os.remove(path)
        # make the index the proper state
        shutil.rmtree(self.index_directory)
        os.mkdir(self.index_directory)
        copy_dir_to_dir(join(".saga/states", state_hash), self.index_directory)

    def set_remote(self, remote_repository):
        with self.remote_location.open("w+") as f:
            f.write(remote_repository)
        print(f"Set new remote repository to {self.remote_repository}")

    def log(self):
        """
        Prints all commits in the current branch to the screen
        """
        for commit_hash in reversed(self.commits[self.head]):
            commit = self.get_commit(commit_hash)
            print("commit: {}".format(commit_hash))
            print("\t{}\n".format(commit.commit_message))

    def status(self):
        """
        Prints the status of the current branch and index
        """
        print("On branch {}".format(self.head))

        rem_state_to_idx, cha_state_to_idx, ins_state_to_idx = changed_files(
            self.curr_state_dir(),
            self.index_directory
        )
        rem_index_to_cur, cha_index_to_cur, ins_index_to_cur = changed_files(
            self.index_directory,
            self.base_directory
        )

        print("Changes staged for commit:")
        for path in rem_state_to_idx:
            print(f"\tremoved: {path}")
        for path in ins_state_to_idx:
            print(f"\tinserted: {path}")
        for path in cha_state_to_idx:
            if path not in cha_index_to_cur and path not in rem_index_to_cur:
                print(f"\tmodified: {path}")

        print("Change not staged for commit:")
        for path in rem_index_to_cur:
            print(f"\tremoved: {path}")
        for path in cha_index_to_cur:
            print(f"\tmodified: {path}")

        print("Untracked files:")
        for path in ins_index_to_cur:
            print(f"\t{path}")

    def diff(self):
        """
        Prints the differences between all files
        """
        _, changed_paths, _ = changed_files(
            self.index_directory,
            self.base_directory
        )

        print("Git diff:")
        operations = []
        for path in changed_paths:
            if os.path.isfile(path):
                print("\tFile:", path)
                file_id = self.file_ids[self.head][path]
                old_file = parse_file(
                    file_id,
                    path,
                    self.index_directory / path
                )
                new_file = parse_file(
                    file_id,
                    path,
                    self.base_directory / path
                )
                ops = old_file.get_operations(new_file)
                for key in ops:
                    print("\t\t:", key, ops[key])
                operations.extend(ops)
            else:
                print("\tDirectory:", path, "changed")

        return operations


import pytest
import os
from tests.cli.cli_utils import saga_folder, run_cmd

def test_init_has_no_remote(saga_folder):
    out = run_cmd("saga remote")
    assert "localhost" in out

def test_set_remote_has_remote(saga_folder):
    out = run_cmd("saga remote testing")
    assert "testing" in out
    out = run_cmd("saga remote")
    assert "testing" in out



import os
import shutil
import requests
from saga.Repository import Repository
from saga.operations.login import login

def push(repository: Repository):
    session = login(f"{repository.remote_repository}/login")

    if session is None:
        print("Login failed. Quiting")
        exit(1)

    url = f"{repository.remote_repository}/cli/push/{repository.base_directory.name}"

    # first we zip the entire repository
    shutil.make_archive("sagatmp", 'zip', repository.base_directory)

    # then we send it across the wire
    with open("sagatmp.zip", 'rb') as f:
        session.post(
            url,
            files={"file": f}
        )
    os.remove("sagatmp.zip")
    print(f"Pushed {repository.base_directory.name} to {repository.remote_repository}")


from os import getcwd
from pathlib import Path
from saga.commands.utils import get_saga_repo_maybe
from saga.operations.init import init


def command_init(args):
    saga_repo = get_saga_repo_maybe()
    if saga_repo is not None:
        print(f"Error: project already exists at {saga_repo.base_directory}")
    else:
        init(Path(getcwd()))



import os
from tests.cli.cli_utils import (
    run_cmd, 
    saga_folder, 
    random_file, 
    current_branch
)


def test_create_branch(saga_folder):
    run_cmd("saga branch newbranch")
    out = run_cmd("saga branch")
    assert "newbranch" in out

def test_switch_branch(saga_folder):
    run_cmd("saga branch newbranch")
    out = run_cmd("saga checkout newbranch")
    print(out)
    out = run_cmd("saga branch")
    out_list = out.split(" ")
    assert "newbranch" in out_list[out_list.index("*") + 1]

def test_cant_switch_uncommitted(saga_folder):
    run_cmd("saga branch newbranch")
    run_cmd("saga checkout newbranch")
    out = run_cmd("saga branch")
    out_list = out.split(" ")
    assert "newbranch" in current_branch()


from saga.Repository import Repository


def list_branches(repository: Repository):
    """
    Lists the branches in the repo, as well as the head
    """
    for branch in repository.branches:
        if branch == repository.head:
            print("   * {}".format(branch))
        else:
            print("     {}".format(branch))



from os import mkdir
from typing import List
from saga.Repository import Repository
from saga.path_utils import copy_dir_to_dir


def commit_int(
        repository: Repository,
        commit_message: str,
        parent_commit_hashes: List[str],
        allow_empty: bool = False
        ):
    """
    Creates a new commit with commit_messages off the parent commit hashes

    TODO: list error conditions
    """
    if not allow_empty and not repository.uncommited_in_index:
        print("Error: no changes to commit")
        return

    # get the information for the commit
    state_hash = repository.index_hash

    # copy the index folder to a new state folder
    state_dir = repository.state_directory / repository.index_hash
    # we don't need to copy if it already exists
    if not state_dir.is_dir():
        mkdir(state_dir)
        copy_dir_to_dir(repository.index_directory, state_dir)

    # add the commit to the commit db
    commit = repository.add_commit(
        state_hash,
        parent_commit_hashes,
        commit_message
    )

    if commit is None:
        print("Error: could not create new commit")
        exit(1)

    # update the head commit of the head branch
    repository.update_head_commit(commit)

    # print that the commit was successful
    print("[{}] {}".format(commit.hash[0:12], commit_message))


def commit(
        repository: Repository,
        commit_message: str,
        allow_empty: bool = False
        ):
    """
    Creates a new commit with commit_messages on the HEAD branch.

    TODO: list error conditions
    """

    # get the information for the commit
    head_commit_hash = repository.head_commit_hash

    commit_int(
        repository,
        commit_message,
        [head_commit_hash],
        allow_empty=allow_empty
    )



import os
from pathlib import Path
from saga.Repository import Repository


def get_saga_repo():
    saga_repo = get_saga_repo_maybe()
    if saga_repo is None:
        print("Error: command cannot run as no saga repo exists")
        exit(1)
    else:
        return saga_repo


def get_saga_repo_maybe():
    # TODO: rewrite with path.parent and Path.home()

    path = os.getcwd()
    while path != "/":
        # check if there is a
        if os.path.isdir(path + "/.saga"):
            return Repository(Path(path))
        path = os.path.dirname(path)
    return None



import os
from pathlib import Path
from tests.cli.cli_utils import run_cmd, saga_folder, random_file
from saga.CommitGraph import CommitGraph
from saga.Repository import Repository

def partial_to_full_hash(saga_folder, partial_hash):
    for path in os.listdir(os.path.join(saga_folder, ".saga", "commits")):
        if path.startswith(partial_hash):
            return path
    return None

def partial_hash_from_blurb(commit_blurb):
    return str(commit_blurb[1:13])

def blurb_to_hash(saga_folder, commit_blurb):
    return partial_to_full_hash(
        saga_folder, 
        partial_hash_from_blurb(commit_blurb)
    )

def test_lca_simple(tmpdir):
    # setup saga and get init hash
    os.chdir(tmpdir)
    init_blurb = run_cmd("saga init")
    init_hash = blurb_to_hash(str(tmpdir), init_blurb)
    # make a branch a and b
    run_cmd("saga branch a")
    run_cmd("saga branch b")
    # make one commit on a
    run_cmd("saga checkout a")
    a_blurb = run_cmd("saga commit --allow-empty -m \"ack\"")
    a_hash = blurb_to_hash(str(tmpdir), a_blurb)
    # make one commit on b
    run_cmd("saga checkout b")
    b_blurb = run_cmd("saga commit --allow-empty -m \"back\"")
    b_hash = blurb_to_hash(str(tmpdir), b_blurb)
    # check that the lca is the initial commit, from both
    repo = Repository(Path(tmpdir))
    commit_graph = CommitGraph(repo)
    need_merge, lcas = commit_graph.least_common_ancestors(a_hash, b_hash)
    assert len(lcas) == 1
    assert lcas.pop() == init_hash


def test_lca_merge(tmpdir):
    # setup saga and get init hash
    os.chdir(tmpdir)
    init_blurb = run_cmd("saga init")
    init_hash = blurb_to_hash(str(tmpdir), init_blurb)
    # make a branch a1, a2, and b
    run_cmd("saga branch a1")
    run_cmd("saga branch a2")
    run_cmd("saga branch b")
    # make one commit on a1 and a2
    run_cmd("saga checkout a1")
    run_cmd("saga commit --allow-empty -m \"ack\"")
    # but we add a random file on a2 so the commit has is different
    run_cmd("saga checkout a2")
    random_file("file")
    run_cmd("saga add file")
    run_cmd("saga commit -m \"ack\"")
    # and then merge them 
    merge_blurb = run_cmd("saga merge a1")
    merge_hash = blurb_to_hash(str(tmpdir), merge_blurb)
    # make one commit on b
    run_cmd("saga checkout b")
    b_blurb = run_cmd("saga commit --allow-empty -m \"back\"")
    b_hash = blurb_to_hash(str(tmpdir), b_blurb)
    # check that the lca is the initial commit, from both
    repo = Repository(Path(tmpdir))
    repo.debug()
    commit_graph = CommitGraph(repo)
    need_merge, lcas = commit_graph.least_common_ancestors(merge_hash, b_hash)
    assert len(lcas) == 1
    assert lcas.pop() == init_hash


def test_lca_two(tmpdir):
    # setup saga and get init hash
    os.chdir(tmpdir)
    run_cmd("saga init")
    # make a branch a and b
    run_cmd("saga branch a")
    run_cmd("saga branch b")
    # make one commit on a
    run_cmd("saga checkout a")
    random_file("filea")
    run_cmd("saga add filea")
    a_blurb = run_cmd("saga commit -m \"ack\"")
    a_commit = blurb_to_hash(str(tmpdir), a_blurb)
    # make one commit on b
    run_cmd("saga checkout b")
    random_file("fileb")
    run_cmd("saga add fileb")
    b_blurb = run_cmd("saga commit -m \"ack\"")
    b_commit = blurb_to_hash(str(tmpdir), b_blurb)
    # merge from b to a
    b_merge_blurb = run_cmd("saga merge a")
    b_merge_hash = blurb_to_hash(str(tmpdir), b_merge_blurb)
    # merge from a to b
    run_cmd("saga checkout a")
    # TODO: we cannot do this currently, as we have no ability to 
    # merge in remote branches -- or merge in by commit hash
    # (which we should have...)
    return
    a_merge_blurb = run_cmd("saga merge b")
    a_merge_hash = blurb_to_hash(str(tmpdir), a_merge_blurb)
    # check that we get two common ancestors
    commit_graph = CommitGraph(os.path.join(tmpdir, ".saga", "commits"))
    lcas = commit_graph.least_common_ancestors(b_merge_hash, a_merge_hash)
    assert len(lcas) == 2
    assert a_commit in lcas and b_commit in lcas





from typing import Optional
from requests import Session
from getpass import getpass

def login(login_url: str) -> Optional[Session]:
    session = Session()

    # TODO: in the future, we can read these
    # from the global saga config file
    username = input("Username: ")
    password = getpass()

    try:
        r = session.post(
            login_url,
            data = {
                "username": username,
                "password": password
            }
        )

        if r.status_code == 200:
            return session
        # login failed
        return None
    except Exception:
        return None



import sys
import os
import pytest
import shlex
from unittest.mock import patch
from saga.main import main
from io import StringIO

def run_cmd(cmd):
    cmd_tokens = shlex.split(cmd)
    # mock the input for the parser
    with patch.object(sys, 'argv', cmd_tokens):
        # capture the output 
        with patch.object(sys, "stdout", StringIO()):
            main()
            out = sys.stdout.getvalue()
            sys.stdout.close()
    return out

@pytest.fixture
def saga_folder(tmpdir):
    os.chdir(tmpdir)
    run_cmd("saga init")
    return tmpdir

def random_file(file_name):
    with open(file_name, 'wb+') as fout:
        fout.write(os.urandom(1024))

def current_branch():
    out = run_cmd("saga branch")
    out_list = out.split(" ")
    return out_list[out_list.index("*") + 1]


import requests
import zipfile
from os import remove

def get_remote(remote_url: str, session=None):
    if session is not None:
        response = session.get(remote_url, stream=True)
    else:
        response = requests.get(remote_url, stream=True)

    # TODO: throw an error, if there is one?
    with open("sagatmp.zip", 'wb') as f:
        for chunk in response.iter_content(chunk_size=8192): 
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)

    with zipfile.ZipFile("sagatmp.zip", 'r') as zip_ref:
        zip_ref.extractall()
    remove("sagatmp.zip")


import os
import pytest
import shutil
import filecmp
from pathlib import Path
from saga.file_types.file_utils import parse_file
from tests.cli.cli_utils import (
    run_cmd, 
    saga_folder, 
    random_file, 
    current_branch
)


def get_test_merges():
    test_merge_names = []
    base = os.path.abspath("tests/merges/")
    for file_type in os.listdir("tests/merges/"):
        merge_file_type = os.path.join(base, file_type)
        if os.path.isdir(merge_file_type):
            for path in os.listdir(merge_file_type):
                path = os.path.join(merge_file_type, path)
                if os.path.isdir(path):
                    test_merge_names.append((path, file_type))

    return test_merge_names
                # we do a merge here


@pytest.mark.parametrize("merge_test_name, file_type", get_test_merges())
def test_merge(saga_folder, merge_test_name, file_type):
    origin = os.path.join(merge_test_name, "origin." + file_type)
    a = os.path.join(merge_test_name, "a." + file_type)
    b = os.path.join(merge_test_name, "b." + file_type)
    merge = os.path.join(merge_test_name, "merge." + file_type)
    file_name = "file." + file_type

    # First, we copy the origin file, and commit it
    shutil.copyfile(origin, saga_folder.join(file_name))
    run_cmd(f"saga add {file_name}")
    run_cmd("saga commit -m \"add file origin\"")

    # then we make branch a and b
    run_cmd("saga branch brancha")
    run_cmd("saga branch branchb")

    # we add file a to branch a
    run_cmd("saga checkout brancha")
    shutil.copyfile(a, file_name)
    run_cmd(f"saga add {file_name}")
    run_cmd("saga commit -m \"add file a\"")

    # we add file b to branch b
    run_cmd("saga checkout branchb")
    shutil.copyfile(b, f"{file_name}")
    run_cmd(f"saga add {file_name}")
    run_cmd("saga commit -m \"add file b\"")

    # then, we try and merge from one branch to the other
    m = run_cmd("saga merge brancha")

    # and check this is the correct merged result
    path = str(saga_folder.join(file_name))
    merge = parse_file("", merge, Path(merge)).file_contents
    assert merge == \
        parse_file("", path, Path(path)).file_contents


import pytest
import os
from tests.cli.cli_utils import run_cmd

def test_init_creates_folder(tmpdir):
    os.chdir(tmpdir)
    out = run_cmd("saga init")
    assert "Created" in out
    assert os.path.exists(tmpdir.join(".saga"))

def test_init_fails_existing(tmpdir):
    os.chdir(tmpdir)
    run_cmd("saga init")
    out = run_cmd("saga init")
    assert "Error" in out

def test_init_fails_existing_superfolder(tmpdir):
    os.chdir(tmpdir)
    run_cmd("saga init")
    subdir = tmpdir.join("/sub")
    os.mkdir(subdir)
    os.chdir(subdir)
    out = run_cmd("saga init")
    assert "Error" in out


from saga.commands.utils import get_saga_repo
from saga.operations.checkout import checkout
from saga.operations.create_branch import create_branch


def command_checkout(args):
    saga_repo = get_saga_repo()
    if args.b:
        create_branch(saga_repo, args.branch)
    checkout(saga_repo, args.branch)



NEED_MERGE = True


class CommitGraph(object):

    def __init__(self, repository):
        self.repository = repository

    def least_common_ancestors(self, commit_hash_1, commit_hash_2):
        # returns a list of all the common ancestors of commits
        lca = set()

        # first, we walk all the ancestors of the first commit hash
        current = {commit_hash_1}
        reached = set()
        while any(current):
            curr = current.pop()
            # subtree
            if curr == commit_hash_2:
                return not NEED_MERGE, {curr}
            if curr not in reached:
                reached.add(curr)
                curr_commit = self.repository.get_commit(curr)
                current.update(set(curr_commit.parent_commit_hashes))

        # then we walk the ancestors of the second commit hash
        current = {commit_hash_2}
        while any(current):
            curr = current.pop()
            # in this case (one branch is subbranch of the other)
            # we can just return this single commit
            if curr == commit_hash_1:
                return not NEED_MERGE, {commit_hash_1}
            if curr in reached:
                lca.add(curr)
            else:
                curr_commit = self.repository.get_commit(curr)
                current.update(set(curr_commit.parent_commit_hashes))

        return NEED_MERGE, lca



from saga.commands.utils import get_saga_repo


def command_remote(args):
    saga_repo = get_saga_repo()
    if args.remote is None:
        print(f"Remote branch is {saga_repo.remote_repository}")
    else:
        saga_repo.set_remote(args.remote)



from pathlib import Path
from shutil import copyfile
from saga.Repository import Repository
from saga.path_utils import copy_dir_to_dir


def add(repository: Repository, add_path: Path):

    add_path = Path(add_path)

    if not add_path.exists():
        print(f"Error: path {add_path} does not exist")
        return

    if add_path.is_file():
        copyfile(add_path, repository.index_directory / add_path)
    else:
        copy_dir_to_dir(
            add_path,
            repository.index_directory,
            exclude=[".saga"]
        )



import requests
import zipfile
from os import remove
from getpass import getpass

def command_test(args):
    s = requests.Session()

    username = input("Username: ")
    password = getpass()

    s.post(
        "http://localhost:3000/login",
        data = {
            "username": username,
            "password": password
        }
    )
    url = "http://localhost:3000/cli/pull/saga"

    with s.get(url, stream=True) as response:
        # throw an error, if there is wone
        response.raise_for_status()
        with open("sagatmp.zip", 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192): 
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)

    with zipfile.ZipFile("sagatmp.zip", 'r') as zip_ref:
        zip_ref.extractall()
    remove("sagatmp.zip")


#!/usr/bin/env python3
from saga.commands.parser import create_saga_parser


def main():
    parser = create_saga_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Optional
from saga.Repository import Repository
from saga.operations.commit import commit


def get_saga_repo_path(directory: Path) -> Optional[Path]:
    while directory != Path("/"):
        if (directory / ".saga").is_dir():
            return directory / ".saga"
        directory = directory.parent
    return None


def init(directory: Path):
    """
    Creates a new saga project in the given directory.

    Fails if the given directory is a sub-directory of an existing
    saga project, or if the directory already contains a saga project.
    """
    # ensure this directory isn't inside saga project
    saga_repo_path = get_saga_repo_path(directory)
    if saga_repo_path is not None:
        print(f"Error: saga project already exists at {saga_repo_path}")

    # create empty saga folder, sub-folders
    (directory / ".saga").mkdir()
    (directory / ".saga" / "commits").mkdir()
    (directory / ".saga" / "states").mkdir()
    (directory / ".saga" / "index").mkdir()
    (directory / ".saga" / "branches").mkdir()

    # create the repository container object
    repository = Repository(directory)

    # create an empty commit on the master branch, so no special cases
    commit(
        repository,
        "Created repository",
        allow_empty=True
    )

    # Notify the user of success
    print(f"Initialized empty saga repository in {repository.saga_directory}")



from tests.cli.cli_utils import run_cmd, saga_folder, random_file

def test_cannot_commit_empty(saga_folder):
    out = run_cmd("saga commit -m \"test\"")
    assert "Error" in out

def test_can_commit_empty_with_flag(saga_folder):
    out = run_cmd("saga commit --allow-empty -m \"test\"")
    assert "Error" not in out


import os
from tests.cli.cli_utils import run_cmd, saga_folder, random_file

def test_add_adds_to_index(saga_folder):
    random_file("file")
    run_cmd("saga add file")
    print(os.listdir())
    assert os.path.exists(saga_folder.join(".saga").join("index").join("file"))

def test_add_updates_index(saga_folder):
    random_file("file")
    run_cmd("saga add file")
    assert os.path.exists(saga_folder.join(".saga").join("index").join("file"))
    random_file("file")
    run_cmd("saga add file")
    import filecmp
    index_file = saga_folder.join(".saga").join("index").join("file")
    filecmp.cmp('file', index_file)


from saga.commands.utils import get_saga_repo
from saga.operations.list_branches import list_branches
from saga.operations.create_branch import create_branch


def command_branch(args):
    saga_repo = get_saga_repo()
    if args.b is not None:
        create_branch(saga_repo, args.b)
    list_branches(saga_repo)



import argparse
from saga.commands.command_add import command_add
from saga.commands.command_branch import command_branch
from saga.commands.command_checkout import command_checkout
from saga.commands.command_clone import command_clone
from saga.commands.command_commit import command_commit
from saga.commands.command_diff import command_diff
from saga.commands.command_init import command_init
from saga.commands.command_log import command_log
from saga.commands.command_merge import command_merge
from saga.commands.command_pull import command_pull
from saga.commands.command_push import command_push
from saga.commands.command_remote import command_remote
from saga.commands.command_status import command_status
from saga.commands.command_test import command_test


def create_saga_parser():
    # create the top-level saga parser
    parser = argparse.ArgumentParser(prog='saga')
    subparsers = parser.add_subparsers(help='sub-command help')

    # create the parser for the "init" command
    parser_init = subparsers.add_parser(
        'init',
        help='creates a new saga project in the current directory'
    )
    parser_init.set_defaults(func=command_init)

    # create the parser for the "add" command
    parser_add = subparsers.add_parser(
        'add',
        help='adds a path to the current index'
    )
    parser_add.add_argument(
        'path',
        type=str,
        help='the path to add to the index'
    )
    parser_add.set_defaults(func=command_add)

    # create the parser for the "commit" command
    parser_commit = subparsers.add_parser(
        'commit',
        help='adds a path to the current index'
    )
    parser_commit.add_argument(
        '--allow-empty',
        action='store_true',
        help='allow the commit to be empty'
    )
    parser_commit.add_argument('-m', type=str, help='commit string')
    parser_commit.set_defaults(func=command_commit)

    # create the parser for the "status" command
    parser_status = subparsers.add_parser(
        'status',
        help='displays info about the current working index'
    )
    parser_status.set_defaults(func=command_status)

    # create the parser for the "log" command
    parser_log = subparsers.add_parser(
        'log',
        help='displays most recent commits'
    )
    parser_log.set_defaults(func=command_log)

    # create the parser for the "diff" command
    parser_diff = subparsers.add_parser(
        'diff',
        help='check the diff of the current index'
    )
    parser_diff.set_defaults(func=command_diff)

    # create the parser for the "branch" command
    parser_branch = subparsers.add_parser(
        'branch',
        help='commands for managing branches'
    )
    parser_branch.add_argument(
        'b',
        nargs="?",
        type=str,
        help='new branch to create'
    )
    parser_branch.set_defaults(func=command_branch)

    # create the parser for the "checkout" command
    parser_checkout = subparsers.add_parser(
        'checkout',
        help='switches the head branch'
    )
    parser_checkout.add_argument(
        '-b',
        help='flag to create a new branch',
        action='store_true'
    )
    parser_checkout.add_argument('branch', type=str, help='branch name')
    parser_checkout.set_defaults(func=command_checkout)

    # create the parser for the "merge" command
    parser_merge = subparsers.add_parser(
        'merge',
        help='merges the head branch with some other branch'
    )
    parser_merge.add_argument(
        'branch',
        type=str,
        help='name of branch to merge'
    )
    parser_merge.set_defaults(func=command_merge)

    # create the parser for the "push" command
    parser_push = subparsers.add_parser(
        'push',
        help='uploads the local repository to the remote repository'
    )
    parser_push.set_defaults(func=command_push)

    # create the parser for the "pull" command
    parser_pull = subparsers.add_parser(
        'pull',
        help='downloads the local repository from the remote repository'
    )
    parser_pull.set_defaults(func=command_pull)

    # create the parser for the "clone" command
    parser_clone = subparsers.add_parser(
        'clone',
        help='downloads a new local repository from the remote repository'
    )
    parser_clone.add_argument(
        'clone_url',
        type=str,
        help='url of remote repository to clone'
    )
    parser_clone.set_defaults(func=command_clone)

    # create the parser for the "remote" command
    parser_remote = subparsers.add_parser(
        'remote',
        help='manage the remote repositories'
    )
    parser_remote.add_argument(
        'remote',
        nargs="?",
        type=str,
        help='repository to track'
    )
    parser_remote.set_defaults(func=command_remote)

    # JUST FOR TESTING!
    parser_test = subparsers.add_parser(
        'test',
    )
    parser_test.set_defaults(func=command_test)

    return parser



from saga.base_file.mixed_data_type.MixedDataType import MixedDataType


class File(object):

    def __init__(
            self,
            file_id,
            file_type,
            file_path,
            file_name,
            file_contents
            ):
        """
        Create a new representation of a file:
            - file_id, file_type are static and can never change
            - file_name, file_contents can
        """
        self.file_id = file_id
        self.file_type = file_type
        self.file_path = file_path
        self.file_name = file_name
        self.file_contents = MixedDataType(file_contents)

    def get_operations(self, other_file):
        """
        Get the operations between this file and the new version
        of the file, other_file
        """
        if self.file_id != other_file.file_id:
            raise ValueError(
                "Can only get operations between \
                    different versions of the same file"
            )

        operations = {"file": []}
        if self.file_name != other_file.file_name:
            operations["file"].append(
                f"Change file name from {self.file_name} \
                    to {other_file.file_name}"
            )

        # now we delegate the operations to the file_contents
        operations.update(
            self.file_contents.get_operations(other_file.file_contents)
        )
        return operations

    # this should be called on the origin file
    def merge(self, file_a, file_b, file_path=None):
        merge_res = self.file_contents.merge(
            file_a.file_contents,
            file_b.file_contents
        )
        if merge_res is None:
            return None
        return File(
            self.file_id,
            self.file_type,
            file_path,
            self.file_name,
            merge_res.mixed_data_type
        )



from saga.Repository import Repository
from saga.path_utils import changed_files


def checkout(repository: Repository, branch_name: str):
    """
    Changes head branch to branch_name.

    Errors if the branch does not exist, or if there
    uncommitted changes in the index, or if there are
    uncommitted changes in the working repository
    to files previously in the state (?)
    """
    if not (repository.branch_directory / branch_name).exists():
        print(f"Error: branch {branch_name} does not exist")
        return

    if repository.uncommited_in_index:
        print(f"Error: uncommitted in index")
        return

    state_path = repository.state_directory / repository.state_hash
    inserted, changed, removed = changed_files(
        repository.base_directory,
        state_path
    )
    if any(changed) or any(removed):
        print("Error: please commit or undo changes before switching branches")
        return

    repository.set_head(branch_name)

    repository.restore_state(repository.state_hash)

    print("Switched to branch {}".format(branch_name))



import os
from tests.cli.cli_utils import (
    run_cmd, 
    saga_folder, 
    random_file, 
    current_branch
)

def test_cannot_checkout_no_branch(saga_folder):
    out = run_cmd("saga checkout newbranch")
    assert "Error" in out

def test_cannot_checkout_create_branch(saga_folder):
    out = run_cmd("saga checkout -b newbranch")
    assert "Error" not in out
    assert "newbranch" in current_branch()

def test_switch_branch_switches_directory(saga_folder):
    run_cmd("saga branch newbranch")
    random_file("file")
    run_cmd("saga add file")
    run_cmd("saga commit -m \"ah\"")
    run_cmd("saga checkout newbranch")
    assert "newbranch" in current_branch()
    assert not os.path.exists(saga_folder.join("file"))

def test_switch_branch_switches_directory_back(saga_folder):
    run_cmd("saga branch newbranch")
    random_file("file")
    run_cmd("saga add file")
    run_cmd("saga commit -m \"ah\"")
    run_cmd("saga checkout newbranch")
    out = run_cmd("saga checkout master")
    assert "master" in current_branch()
    assert os.path.exists(saga_folder.join("file"))

def test_cannot_switch_removed_change(saga_folder):
    run_cmd("saga branch newbranch")
    random_file("file")
    run_cmd("saga add file")
    run_cmd("saga commit -m \"ack\"")
    os.remove("file")
    out = run_cmd("saga checkout newbranch")
    assert "Error" in out

def test_cannot_switch_uncommitted_changes(saga_folder):
    run_cmd("saga branch newbranch")
    random_file("file")
    run_cmd("saga add file")
    out = run_cmd("saga checkout newbranch")
    assert "Error" in out